/*
 * Copyright (c) 2010, Tess Snider

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
*/

using Gtk;
using System;
using System.IO;
using System.IO.Compression;
using System.Collections.Generic;
using System.Text.RegularExpressions;

namespace DesertPaintLab
{
    public class PlayerProfile
    {
        public string Name
        {
            get; private set;
        }
        string directory;
        string reactFile;
        string reagentFile;
        string settingsFile;

        static Regex recipeHeaderRegex = new Regex(@"^--- Recipe: (?<colorname>(\w*\s)*\w+)\s*");
        static Regex recipeIngredientRegex = new Regex(@"(?<ingredient>(\w+\s)?\w+)\s*\|\s*(?<quantity>\d+)\s*");

        ReactionSet reactions = new ReactionSet();
        // ingredient -> [ingredient, reaction]
        //SortedDictionary<string, SortedDictionary<string, Reaction>> reactions =
        //	new SortedDictionary<string, SortedDictionary<string, Reaction>>();
        SortedDictionary<string, PaintRecipe> recipes;
        SortedDictionary<string, PaintRecipe> ribbonRecipes;

        Settings settings = new Settings();
        public Settings ProfileSettings
        {
            get
            {
                return settings;
            }
        }

        static PlayerProfile current = null;

        static PlayerProfile Current
        {
            get
            {
                return current;
            }
        }

        public string LastError
        {
            get; private set;
        }

        public PlayerProfile(string name, string directory)
        {
            this.Name = name;
            this.directory = directory;
            this.reactFile = System.IO.Path.Combine(directory, "dp_reactions.txt");
            this.reagentFile = System.IO.Path.Combine(directory, "ingredients.txt");
            this.settingsFile = System.IO.Path.Combine(directory, "settings");
            this.recipes = new SortedDictionary<string, PaintRecipe>();
            this.ribbonRecipes = new SortedDictionary<string, PaintRecipe>();
            foreach (PaintColor color in Palette.Colors)
            {
                this.recipes.Add(color.Name, new PaintRecipe());
            }
            foreach (PaintColor color in Palette.Colors)
            {
                this.ribbonRecipes.Add(color.Name, new PaintRecipe());
            }
        }

        public string Directory
        {
            get
            {
                return this.directory;
            }
        }

        public ReactionSet Reactions
        {
            get
            {
                return this.reactions;
            }
        }

        public string ReagentFile
        {
            get
            {
                return this.reagentFile;
            }
        }

        public SortedDictionary<string, PaintRecipe> Recipes
        {
            get
            {
                return this.recipes;
            }
        }

        public SortedDictionary<string, PaintRecipe> RibbonRecipes
        {
            get
            {
                return this.ribbonRecipes;
            }
        }

        public int RecipeCount
        {
            get
            {
                int count = 0;
                foreach (PaintRecipe recipe in this.recipes.Values)
                {
                    if (recipe.IsValidForConcentration(PaintRecipe.PAINT_RECIPE_MIN_CONCENTRATION))
                    {
                        ++count;
                    }
                }
                return count;
            }
        }

        public int RibbonCount
        {
            get
            {
                int count = 0;
                foreach (PaintRecipe recipe in this.ribbonRecipes.Values)
                {
                    if (recipe.IsValidForConcentration(PaintRecipe.RIBBON_RECIPE_MIN_CONCENTRATION))
                    {
                        ++count;
                    }
                }
                return count;
            }
        }

        public bool Initialize()
        {
            // Copy template files into new directory.
            string templatePath = FileUtils.FindApplicationResourceDirectory("template");

            if (!System.IO.Directory.Exists(templatePath))
            {
                LastError = "Failed to find profile template folder.";
                return false;
            }

            // Create new directory.
            System.IO.Directory.CreateDirectory(directory);

            DirectoryInfo di = new DirectoryInfo(templatePath);
            FileInfo[] templateFiles = di.GetFiles();

            foreach (FileInfo file in templateFiles)
            {
                string destFile = System.IO.Path.Combine(directory, file.Name);
                System.IO.File.Copy(file.FullName, destFile, true);
                if (!File.Exists(destFile))
                {
                    LastError = "Failed to copy template file " + file.Name + ".";
                    return false;
                }
            }
            return true;
        }

        private void WriteReaction(StreamWriter writer, string reagent1, string reagent2, string r, string g, string b)
        {
            writer.Write(reagent1);
            writer.Write(" ");
            writer.Write(reagent2);
            writer.Write(" ");
            writer.Write(r);
            writer.Write(" ");
            writer.Write(g);
            writer.Write(" ");
            writer.WriteLine(b);
        }

        public void ConvertFromPP(string ppFile, string dpFile)
        {
            string line;
            using (StreamReader reader = new StreamReader(ppFile))
            {
                using (StreamWriter writer = new StreamWriter(dpFile))
                {
                    while ((line = reader.ReadLine()) != null)
                    {
                        string[] tokens = line.Split('|');
                        //if ((tokens.Length > 0) && (tokens [0] != "//"))
                        if ((tokens.Length != 5) && (tokens[0].Trim() != "//"))
                        {
                            string reagent1 = tokens[0].Trim();
                            string reagent2 = tokens[1].Trim();
                            string colorCode = tokens[2].Trim();
                            string change1 = tokens[3].Trim();
                            string change2 = tokens[4].Trim();
                            // Write reaction.
                            switch (colorCode)
                            {
                            case "W":
                                WriteReaction(writer, reagent1, reagent2, change1, change1, change1);
                                WriteReaction(writer, reagent2, reagent1, change2, change2, change2);
                                break;
                            case "R":
                                WriteReaction(writer, reagent1, reagent2, change1, "0", "0");
                                WriteReaction(writer, reagent2, reagent1, change2, "0", "0");
                                break;
                            case "G":
                                WriteReaction(writer, reagent1, reagent2, "0", change1, "0");
                                WriteReaction(writer, reagent2, reagent1, "0", change2, "0");
                                break;
                            case "B":
                                WriteReaction(writer, reagent1, reagent2, "0", "0", change1);
                                WriteReaction(writer, reagent2, reagent1, "0", "0", change2);
                                break;
                            }
                        }
                    }
                }
            }
        }

        public bool SaveToPP(string ppFile)
        {
            Reaction reaction1, reaction2;
            using (StreamWriter writer = new StreamWriter(ppFile))
            {
                foreach (string reagentName1 in ReagentManager.Names)
                {
                    // TODO: could be more efficient by only iterating over the names after reagent1
                    foreach (string reagentName2 in ReagentManager.Names)
                    {
                        if (reagentName1.Equals(reagentName2))
                        {
                            continue;
                        }
                        Reagent reagent1 = ReagentManager.GetReagent(reagentName1);
                        Reagent reagent2 = ReagentManager.GetReagent(reagentName2);
                        reaction1 = reactions.Find(reagent1, reagent2);
                        if (reaction1 != null && !reaction1.Exported)
                        {
                            reaction2 = reactions.Find(reagent2, reagent1);
                            if (reaction2 != null)
                            {
                                writer.Write(reagent1.PracticalPaintName + " | " + reagent2.PracticalPaintName + " | ");
                                if ((Math.Abs(reaction1.Red) > Math.Abs(reaction1.Green)) ||
                                    (Math.Abs(reaction2.Red) > Math.Abs(reaction2.Green)))
                                {
                                    writer.WriteLine("R | " + reaction1.Red + " | " + reaction2.Red);
                                }
                                else if ((Math.Abs(reaction1.Green) > Math.Abs(reaction1.Red)) ||
                                    (Math.Abs(reaction2.Green) > Math.Abs(reaction2.Red)))
                                {
                                    writer.WriteLine("G | " + reaction1.Green + " | " + reaction2.Green);
                                }
                                else if ((Math.Abs(reaction1.Blue) > Math.Abs(reaction1.Red)) ||
                                    (Math.Abs(reaction2.Blue) > Math.Abs(reaction2.Red)))
                                {
                                    writer.WriteLine("B | " + reaction1.Blue + " | " + reaction2.Blue);
                                }
                                else
                                {
                                    writer.WriteLine("W | " + reaction1.Red + " | " + reaction2.Red);
                                }
                                reaction1.Exported = true;
                                reaction2.Exported = true;
                            }
                        }
                    }
                }
            }

            // Clear Exported flags.
            foreach (string reagentName1 in ReagentManager.Names)
            {
                // TODO: could be more efficient by only iterating over the names after reagent1
                foreach (string reagentName2 in ReagentManager.Names)
                {
                    if (reagentName1.Equals(reagentName2))
                    {
                        continue;
                    }
                    Reagent reagent1 = ReagentManager.GetReagent(reagentName1);
                    Reagent reagent2 = ReagentManager.GetReagent(reagentName2);
                    reaction1 = reactions.Find(reagent1, reagent2);
                    if (reaction1 != null)
                    {
                        reaction1.Exported = false;
                    }
                    reaction2 = reactions.Find(reagent2, reagent1);
                    if (reaction2 != null)
                    {
                        reaction2.Exported = false;
                    }
                }
            }
            return true;
        }

        public void ImportFromPP(string importDir)
        {
            // Convert old file.
            ConvertFromPP(
                System.IO.Path.Combine(importDir, "reactions.txt"),
                reactFile);
            try
            {
                // If there is an ingredients file, move it in.
                System.IO.File.Copy(
                    System.IO.Path.Combine(importDir, "ingredients.txt"),
                    System.IO.Path.Combine(directory, "ingredients.txt"),
                    true);
            }
            catch (Exception)
            {
                // If there is no ingredients file, we don't really care.	
            }
        }

        public void Import(string file)
        {
            ZipFile.ExtractToDirectory(file, directory);
        }

        public void Export(string file)
        {
            ZipFile.CreateFromDirectory(directory, file);
        }

        public bool Load()
        {
            string line;
            settings.Reset();
            settings.Load(settingsFile);
            reactions.Clear();
            if (File.Exists(reagentFile))
            {
                ReagentManager.LoadProfileReagents(reagentFile);
            }
            else
            {
                LastError = "Failed to find profile reagents file.";
                return false;
            }
            ReagentManager.InitializeReactions(ref reactions);
            if (!File.Exists(reactFile))
            {
                LastError = "Failed to find profile reactions file.";
                return false;
            }
            using (StreamReader reader = new StreamReader(reactFile))
            {
                while ((line = reader.ReadLine()) != null)
                {
                    string[] tokens = line.Split(' ');
                    if (tokens.Length == 5)
                    {
                        Reagent reagent1 = ReagentManager.GetReagent(tokens[0].Trim());
                        Reagent reagent2 = ReagentManager.GetReagent(tokens[1].Trim());
                        Reaction reaction = new Reaction(
                            int.Parse(tokens[2].Trim()),
                            int.Parse(tokens[3].Trim()),
                            int.Parse(tokens[4].Trim())
                            );
                        reactions.Set(reagent1, reagent2, reaction);
                    }
                }
            }
            return true;
        }

        public void Save()
        {
            settings.Save(settingsFile);
            Reaction reaction;
            using (StreamWriter writer = new StreamWriter(reactFile, false))
            {
                foreach (string reagentName1 in ReagentManager.Names)
                {
                    // TODO: could be more efficient by only iterating over the names after reagent1
                    foreach (string reagentName2 in ReagentManager.Names)
                    {
                        if (reagentName1.Equals(reagentName2))
                        {
                            continue;
                        }
                        Reagent reagent1 = ReagentManager.GetReagent(reagentName1);
                        Reagent reagent2 = ReagentManager.GetReagent(reagentName2);
                        reaction = reactions.Find(reagent1, reagent2);
                        if (reaction != null)
                        {
                            writer.WriteLine(reagent1.PracticalPaintName + " " + reagent2.PracticalPaintName + " " +
                            reaction.Red + " " + reaction.Green + " " + reaction.Blue);
                        }
                    }
                }
            }
        }

        private void LoadRecipes(SortedDictionary<string, PaintRecipe> recipeDict, string filename, uint concentration)
        {
            foreach (PaintRecipe recipe in recipeDict.Values)
            {
                recipe.Clear();
            }
            string recipeFile = System.IO.Path.Combine(directory, filename);
            string line;
            Match match;
            bool inRecipe = false;
            PaintRecipe testRecipe = new PaintRecipe();
            testRecipe.Reactions = reactions;
            string currentRecipeColor = null;
            if (File.Exists(recipeFile))
            {
                using (StreamReader reader = new StreamReader(recipeFile))
                {
                    while ((line = reader.ReadLine()) != null)
                    {
                        match = recipeHeaderRegex.Match(line);
                        if (match.Success)
                        {
                            if (testRecipe != null && currentRecipeColor != null)
                            {
                                if (testRecipe.IsValidForConcentration(concentration))
                                {
                                    recipeDict[currentRecipeColor].CopyFrom(testRecipe);
                                }
                            }
                            testRecipe.Clear();
                            currentRecipeColor = match.Groups["colorname"].Value;
                            inRecipe = true;
                        }
                        else if (inRecipe)
                        {
                            match = recipeIngredientRegex.Match(line);
                            if (match.Success)
                            {
                                string ingredient = match.Groups["ingredient"].Value;
                                uint quantity = uint.Parse(match.Groups["quantity"].Value);
                                testRecipe.AddReagent(ingredient, quantity);
                            }
                        }
                    }
                    if (inRecipe)
                    {
                        if (testRecipe.IsValidForConcentration(concentration))
                        {
                            recipeDict[currentRecipeColor].CopyFrom(testRecipe);
                        }
                    }
                }
            }
        }

        private void SaveRecipes(SortedDictionary<string, PaintRecipe> recipeDict, string filename)
        {
            if (recipeDict != null)
            {
                string recipeFile = System.IO.Path.Combine(directory, filename);
                using (StreamWriter writer = new StreamWriter(recipeFile, false))
                {
                    foreach (KeyValuePair<string, PaintRecipe> pair in recipeDict)
                    {
                        writer.WriteLine("--- Recipe: {0}", pair.Key);
                        foreach (PaintRecipe.RecipeIngredient ingredient in pair.Value.Ingredients)
                        {
                            writer.WriteLine("{0,-14} | {1}", ingredient.name, ingredient.quantity);
                        }
                    }
                }
            }
        }

        public void LoadRecipes()
        {
            LoadRecipes(this.recipes, "dp_recipes.txt", PaintRecipe.PAINT_RECIPE_MIN_CONCENTRATION);
            LoadRecipes(this.ribbonRecipes, "dp_ribbons.txt", PaintRecipe.RIBBON_RECIPE_MIN_CONCENTRATION);
        }

        public void SaveRecipes()
        {
            SaveRecipes(this.recipes, "dp_recipes.txt");
            SaveRecipes(this.ribbonRecipes, "dp_ribbons.txt");
        }

        public void ExportWikiRecipes(string file)
        {
            StreamWriter writer = new StreamWriter(file);
            ExportWikiFormat(writer, this.recipes);
        }

        public void ExportWikiRibbons(string file)
        {
            StreamWriter writer = new StreamWriter(file);
            ExportWikiFormat(writer, this.ribbonRecipes);
        }

        public void ExportWikiRecipes(TextWriter writer)
        {
            ExportWikiFormat(writer, this.recipes);
        }

        public void ExportWikiRibbons(TextWriter writer)
        {
            ExportWikiFormat(writer, this.ribbonRecipes);
        }

        static public PaintColor RgbToHsv(PaintColor color)
        {
            float h = 0;
            float s = 0;
            float r = color.Red / 255.0f;
            float g = color.Green / 255.0f;
            float b = color.Blue / 255.0f;
            float cmax = Math.Max(Math.Max(r, g), b);
            float cmin = Math.Min(Math.Min(r, g), b);
            float cd = cmax - cmin;
            if (cd == 0)
            {
                h = 0;
            }
            else if (cmax == r)
            {
                h = (((g - b) / cd) % 6) / 6f;
            }
            else if (cmax == g)
            {
                h = (((b - r) / cd) + 2) / 6f;
            }
            else
            {
                h = (((r - g) / cd) + 4) / 6f;
            }

            if (cmax == 0)
            {
                s = 0;
            }
            else
            {
                s = cd / cmax;
            }

            if (h < 0)
            {
                h = 1 + h;
            }

            if (h >= 1)
            {
                h = 0;
            }


            return new PaintColor((byte)(h * 255), (byte)(s * 255), (byte)(cmax * 255));
        }

        public void ExportWikiFormat(TextWriter writer, SortedDictionary<string, PaintRecipe> recipeDict)
        {
            PaintRecipe recipe;
            List<KeyValuePair<string, string>> missing = new List<KeyValuePair<string, string>>();
            using (writer)
            {
                writer.WriteLine("{| class='wikitable sortable' border=\"1\" style=\"background-color:#DEB887;\"");
                writer.WriteLine("! Color !! Recipe !! Missing Reactions? || Verified");
                foreach (PaintColor color in Palette.Colors)
                {
                    writer.WriteLine("|-");
                    string colorLine = "| ";
                    colorLine += "style=\"font-weight: bold; background-color: #" + color.Red.ToString("X2") + color.Green.ToString("X2") + color.Blue.ToString("X2") + ";";
                    PaintColor hsvColor = RgbToHsv(color);
                    if (hsvColor.Blue < 200) // value
                    {
                        // dark color gets light text
                        colorLine += " color: #FFFFFF;";
                    }
                    else
                    {
                        colorLine += "color: #000000;";
                    }
                    colorLine += "\" | " + color.Name + " || ";
                    if (recipeDict.TryGetValue(color.Name, out recipe))
                    {
                        foreach (PaintRecipe.RecipeIngredient ingredient in recipe.Ingredients)
                        {
                            colorLine += " " + ingredient.ToString();
                        }
                    }
                    else
                    {
                        // no recipe
                    }
                    colorLine += " || ";

                    if (recipe.CheckMissingReactions(ref missing) == true)
                    {
                        colorLine += "Y";
                    }
                    else
                    {
                        colorLine += "N";
                    }

                    colorLine += " || N";
                    writer.WriteLine(colorLine);
                }
                writer.WriteLine("|}");
            }
        }

        public Reaction FindReaction(Reagent reagent1, Reagent reagent2)
        {
            return reactions.Find(reagent1, reagent2);
        }

        public void SetReaction(Reagent reagent1, Reagent reagent2, Reaction reaction)
        {
            reactions.Set(reagent1, reagent2, reaction);
        }

        public void ClearReaction(Reagent reagent1, Reagent reagent2)
        {
            reactions.Remove(reagent1, reagent2);
        }

        public void SetRecipe(PaintRecipe recipe)
        {
            string colorName = Palette.FindNearest(recipe.ReactedColor);
            recipes[colorName].CopyFrom(recipe);
        }

        public void SetRibbonRecipe(PaintRecipe recipe)
        {
            string colorName = Palette.FindNearest(recipe.ReactedColor);
            ribbonRecipes[colorName].CopyFrom(recipe);
        }

    }
}

