﻿/*
 * Copyright (c) 2015, Jason Maltzen

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
*/

using System;
using System.IO;

namespace DesertPaintLab
{
    // ReactionRecorder - business logic for recording paint reactions
    public class ReactionRecorder
    {
        public delegate void CaptureProgressHandler(int x, int y);
        public CaptureProgressHandler OnCaptureProgress;

        public enum LogVerbosity
        {
            Low,
            Normal,
            High,
            Excessive
        }

        const int COLOR_TOLERANCE = 3;

        const InterfaceSize DEFAULT_INTERFACE_SIZE = InterfaceSize.Small;

        // Swatch is 301x20 solid color, 1px darker left/top, 1px black outer left/top, 1px light right/bottom, 1px bright right/bottom
        // top-right and bottom-left are bright instead of black

        // Color bars are 4x302 solid color, 2px border on all sides (darker left/top, lighter bottom/right

        public static readonly int[] SWATCH_HEIGHT = 
        {
            24, // tiny
            24, // small
            24, // medium
            24, // large
            24 // huge
        };
        public static readonly int[] SWATCH_WIDTH =
        {
            306, // tiny
            306, // small
            306, // medium
            320, // large
            350 // huge
        };
        public static readonly int[] COLOR_BAR_WIDTH =
        {
            306, // tiny
            306, // small -- includes left and right borders
            306, // medium
            320, // large
            350 // huge
        };
        public static readonly int[] RED_BAR_SPACING =
        {
            32, // tiny
            32, // small
            32, // medium
            32, // large
            32 // huge
        };
        public static readonly int[] GREEN_BAR_SPACING =
        {
            42, // tiny
            42, // small
            42, // medium
            42, // large
            42 // huge
        };
        public static readonly int[] BLUE_BAR_SPACING =
        {
            52, // tiny
            52, // small
            52, // medium
            52, // large
            52 // huge
        };
        // width to test on ends of swatch (10% on either end)
        public static readonly int[] SWATCH_TEST_WIDTH =
        {
            26, // tiny
            26, // small
            26, // medium
            28, // large
            31  // huge
        };

        int swatchHeight = SWATCH_HEIGHT[(int)DEFAULT_INTERFACE_SIZE];
        int swatchWidth = SWATCH_WIDTH[(int)DEFAULT_INTERFACE_SIZE];
        int swatchTestWidth = SWATCH_TEST_WIDTH[(int)DEFAULT_INTERFACE_SIZE];
        int colorBarWidth = COLOR_BAR_WIDTH[(int)DEFAULT_INTERFACE_SIZE];
        int redBarSpacing = RED_BAR_SPACING[(int)DEFAULT_INTERFACE_SIZE];
        int greenBarSpacing = GREEN_BAR_SPACING[(int)DEFAULT_INTERFACE_SIZE];
        int blueBarSpacing = BLUE_BAR_SPACING[(int)DEFAULT_INTERFACE_SIZE];

        public int SwatchHeight {  get { return swatchHeight; } }
        public int SwatchWidth {  get { return swatchWidth; } }
        public int ColorBarWidth { get { return colorBarWidth;  } }
        public int RedBarSpacing {  get { return redBarSpacing; } }
        public int GreenBarSpacing { get { return greenBarSpacing; } }
        public int BlueBarSpacing {  get { return blueBarSpacing; } }

        // Current Status
        public bool IsCaptured { get; private set; }
        public int X { get; private set; }
        public int Y { get; private set; }
        public int ScreenWidth { get; private set; }
        public int ScreenHeight { get; private set; }
        private PaintColor _recordedColor = new PaintColor();
        public PaintColor RecordedColor { get { return _recordedColor; } private set { _recordedColor.Set(value); } }
        public int RedBarX { get; private set; }
        public int RedBarY { get; private set; }

        int pixelMultiplier = 1;
        public int PixelMultiplier
        {
            get
            {
                return pixelMultiplier;
            }
            set
            {
                pixelMultiplier = value;
                UpdateSwatchSizes();
            }
        }
        private InterfaceSize interfaceSize;
        public InterfaceSize InterfaceSize
        {
            get
            {
                return interfaceSize;
            }
            set
            {
                interfaceSize = value;
                UpdateSwatchSizes();
            }
        }

        bool firstRun = true;
        int lastSwatchX = -1;
        int lastSwatchY = -1;
        public int SwatchX { get { return lastSwatchX; } }
        public int SwatchY { get { return lastSwatchY; } }

        private LogVerbosity logVerbosity = LogVerbosity.Normal;

        private class PixelColor
        {
            public byte r;
            public byte g;
            public byte b;

            public bool IsMatch(PixelColor other)
            {
                return ((Math.Abs(r - other.r) <= COLOR_TOLERANCE) &&
                    (Math.Abs(g - other.g) <= COLOR_TOLERANCE) &&
                    (Math.Abs(b - other.b) <= COLOR_TOLERANCE));
            }

            public static bool IsDark(PixelColor color)
            {
                return (color.r < 0x47) && (color.g < 0x47) && (color.b < 0x47);
            }

            public static bool IsPapyrus(PixelColor color)
            {
                // red between 208 and 244
                // 240 and 255
                // green between 192 and 237
                // 223 and 250
                // blue between 145 and 205
                // 178 and 232
                //return ((r > 0xD0) && (g >= 0xC0) && (b >= 0x91)) &&
                //       ((r < 0xF4) && (g <= 0xED) && (b <= 0xCD));
                return ((color.r >= 0xF0) && (color.r <= 0xFF) && (color.g >= 0xDF) && (color.g <= 0xFA) && (color.b >= 0xB2) && (color.b <= 0xE8));
            }

            public static bool IsRed(PixelColor color)
            {
                return (color.r > 0x9F) && (color.g < 0x70) && (color.b < 0x70);
            }
            public static bool IsGreen(PixelColor color)
            {
                return (color.r < 0x70) && (color.g > 0x9F) && (color.b < 0x70);
            }
            public static bool IsBlue(PixelColor color)
            {
                return (color.r < 0x70) && (color.g < 0x70) && (color.b > 0x9F);
            }
        }

        unsafe private class Pixels
        {
            public byte* pixBytes;
            public int width;
            public int height;
            public int stride;
            public int pixelMultiplier;
            public int numChannels;

            private PixelColor _tempColor = new PixelColor();
            private PixelColor _tempColor2 = new PixelColor();

            public int ComputeOffset(int x, int y)
            {
                return (y * pixelMultiplier * stride) + (x * pixelMultiplier * numChannels);
            }

            public int UpdateOffset(int offset, int xChange, int yChange)
            {
                return offset + (yChange * pixelMultiplier * stride) + (xChange * pixelMultiplier * numChannels);
            }

            unsafe public void ColorAt(int offset, ref PixelColor pixel)
            {
                pixel.r = pixBytes[offset];
                pixel.g = pixBytes[offset + 1];
                pixel.b = pixBytes[offset + 2];
            }

            unsafe public void ColorAt(int x, int y, ref PixelColor pixel)
            {
                int offset = ComputeOffset(x, y);
                ColorAt(offset, ref pixel);
            }

            unsafe public bool DoesPixelMatch(int x, int y, Func<PixelColor, bool> matchFunc)
            {
                int offset = ComputeOffset(x, y);
                ColorAt(offset, ref _tempColor);
                return matchFunc(_tempColor);
            }

            unsafe public bool DoesPixelMatch(int offset, Func<PixelColor, bool> matchFunc)
            {
                ColorAt(offset, ref _tempColor);
                return matchFunc(_tempColor);
            }

            // Compute length of horizontal bar starting at x,y using matching function
            unsafe public int LengthOfColorAt(int x, int y, Func<PixelColor, bool> matchesColor)
            {
                int count = 0;
                for (int xVal = x; xVal < width; ++xVal)
                {
                    int offset = ComputeOffset(xVal, y);
                    if (!DoesPixelMatch(xVal, y, matchesColor)) break;
                    ++count;
                }
                return count;
            }

            unsafe public bool IsSolidPatchAt(int x, int y, int patchWidth, int patchHeight)
            {
                if ((x + patchWidth >= width) || (y + patchHeight >= height)) return false;
                ColorAt(x, y, ref _tempColor2);
                Console.WriteLine("color at {0},{1} = {2},{3},{4}", x, y, _tempColor2.r, _tempColor2.g, _tempColor2.b);
                bool ok = true;
                ok &= DoesPixelMatch(x + patchWidth - 1, y, _tempColor2.IsMatch);
                ok &= DoesPixelMatch(x, y + patchHeight - 1, _tempColor2.IsMatch);
                ok &= DoesPixelMatch(x + patchWidth - 1, y + patchHeight - 1, _tempColor2.IsMatch);
                for (int xOff = 1; ok && (xOff < patchWidth - 1); ++xOff)
                {
                    ok &= DoesPixelMatch(x + xOff, y, _tempColor2.IsMatch);
                    ok &= DoesPixelMatch(x + xOff, y + patchHeight - 1, _tempColor2.IsMatch);
                }
                for (int yOff = 1; ok && (yOff < patchHeight - 1); ++yOff)
                {
                    ok &= DoesPixelMatch(x, y + yOff, _tempColor2.IsMatch);
                    ok &= DoesPixelMatch(x + patchWidth - 1, y + yOff, _tempColor2.IsMatch);
                }
                for (int yOff = 1; ok && (yOff < patchHeight - 1); ++yOff)
                {
                    for (int xOff = 1; ok && (xOff < patchWidth - 1); ++xOff)
                    {
                        ok &= DoesPixelMatch(x + xOff, y + yOff, _tempColor2.IsMatch);
                    }
                }
                return ok;
            }
        }

        private static ReactionRecorder _instance;
        public static ReactionRecorder Instance
        {
            get {
                if (_instance == null)
                {
                    _instance = new ReactionRecorder();
                }
                return _instance;
            }
        }

        public StreamWriter Log { get; set; }
        private void WriteLog(LogVerbosity verbosity, string format, params object[] args)
        {
            if ((Log != null) && (verbosity <= logVerbosity))
            {
                Log.WriteLine(format, args);
            }
        }

        public delegate void ReactionRecordedHandler(Reagent reagent1, Reagent reagent2, Reaction reaction);
        public ReactionRecordedHandler OnReactionRecorded;

        public ReactionRecorder()
        {
            pixelMultiplier = 1;
            interfaceSize = DEFAULT_INTERFACE_SIZE;

            UpdateSwatchSizes();
        }

        public ReactionRecorder(int pixelMultiplier)
        {
            this.pixelMultiplier = pixelMultiplier;
            this.interfaceSize = DEFAULT_INTERFACE_SIZE;

            UpdateSwatchSizes();
        }

        public ReactionRecorder(int pixelMultiplier, InterfaceSize interfaceSize)
        {
            this.pixelMultiplier = pixelMultiplier;
            this.interfaceSize = interfaceSize;

            UpdateSwatchSizes();
        }

        public void SetPixelMultiplier(int pixelMultiplier)
        {
            this.pixelMultiplier = pixelMultiplier;

            UpdateSwatchSizes();
        }

        public void SetInterfaceSize(InterfaceSize interfaceSize)
        {
            this.interfaceSize = interfaceSize;

            UpdateSwatchSizes();
        }

        private void UpdateSwatchSizes()
        {
            swatchHeight = SWATCH_HEIGHT[(int)interfaceSize] * pixelMultiplier;
            swatchWidth = SWATCH_WIDTH[(int)interfaceSize] * pixelMultiplier;
            colorBarWidth = COLOR_BAR_WIDTH[(int)interfaceSize] * pixelMultiplier;
            redBarSpacing = RED_BAR_SPACING[(int)interfaceSize] * pixelMultiplier;
            greenBarSpacing = GREEN_BAR_SPACING[(int)interfaceSize] * pixelMultiplier;
            blueBarSpacing = BLUE_BAR_SPACING[(int)interfaceSize] * pixelMultiplier;
            swatchTestWidth = SWATCH_TEST_WIDTH[(int)interfaceSize] * pixelMultiplier;
        }

        unsafe private bool IsPossibleSwatchSlice(Pixels pixels, int x, int y)
        {
            int testPixelStart = pixels.ComputeOffset(x, y);
            PixelColor color = new PixelColor();
            pixels.ColorAt(x, y, ref color);

            bool isDarkPixel = false;
            //bool isPapyAbove = false;
            //bool isPapyBelow = false;

            // 1.) Check if the top pixel is a dark pixel.
            isDarkPixel = PixelColor.IsDark(color);
            
            //// 2.) Check the pixel above it to see if it's from the papy texture.
            //int otherPixelStart = testPixelStart - stride;
            //isPapyAbove = ((otherPixelStart >= 0) &&
            //    IsPapyTexture(pixBytes[otherPixelStart++], pixBytes[otherPixelStart++], pixBytes[otherPixelStart]));

            //// 3.) Check the pixel below where the swatch should be to see if it's also from the papy texture.
            //otherPixelStart = testPixelStart + (stride * swatchHeight);
            //isPapyBelow = (IsPapyTexture(pixBytes[otherPixelStart++], pixBytes[otherPixelStart++], pixBytes[otherPixelStart]));

            //bool result = isDarkPixel && isPapyAbove && isPapyBelow;
            bool result = isDarkPixel;

            // grab the swatch color
            // 2 down from top border
            pixels.ColorAt(x, y + 2, ref color);

            // scan the column from 2 below the top to 3 above the bottom to ensure the color matches
            for (int i = 2; result && (i < (swatchHeight-3)); ++i)
            {
                result &= pixels.DoesPixelMatch(x, y + i, color.IsMatch);
            }

            if (!result)
            {
                WriteLog(LogVerbosity.Normal, "Swatch slice at {0}, {1} failed to match", x, y);
            }

            return result;
        }

        unsafe private bool IsPossibleSwatchUpperLeft(Pixels pixels, int x, int y)
        {
            int testPixelStart = pixels.ComputeOffset(x, y);

            if (testPixelStart < pixels.stride)
            {
                WriteLog(LogVerbosity.Normal, "Can't test {0},{1} - is not at least 1 row in", x, y);
                return false;
            }

            bool result = true;

            int swatchSolidWidth = swatchWidth - 4;
            int swatchSolidHeight = swatchHeight - 5; // 2 top and 3 bottom pixels are slightly different colors
            int swatchSolidLeftX = x + 2;
            int swatchSolidTopY = y + 2;
            int swatchSolidRightX = swatchSolidLeftX + swatchSolidWidth - 1;
            int swatchSolidBottomY = swatchSolidTopY + swatchSolidHeight - 1;

            PixelColor swatchColor = new PixelColor();
            pixels.ColorAt(swatchSolidLeftX, swatchSolidTopY, ref swatchColor);

            // Check the other 3 corners of the swatch size for color match
            PixelColor testColor = new PixelColor();
            bool upperRightResult = pixels.DoesPixelMatch(swatchSolidRightX, swatchSolidTopY, swatchColor.IsMatch);
            if (!upperRightResult)
            {
                pixels.ColorAt(swatchSolidRightX, swatchSolidTopY, ref testColor);
                WriteLog(LogVerbosity.Excessive, "Upper-right mismatch for {8}, {9} - found {0},{1},{2} at {3}, {4} expected {5},{6},{7}", testColor.r, testColor.g, testColor.b, swatchSolidRightX, swatchSolidTopY, swatchColor.r, swatchColor.g, swatchColor.b, x, y);
            }
            bool lowerLeftResult = pixels.DoesPixelMatch(swatchSolidLeftX, swatchSolidBottomY, swatchColor.IsMatch);
            if (!lowerLeftResult)
            {
                pixels.ColorAt(swatchSolidLeftX, swatchSolidBottomY, ref testColor);
                WriteLog(LogVerbosity.Excessive, "Lower-left mismatch for {8}, {9} - found {0},{1},{2} at {3}, {4} expected {5},{6},{7}", testColor.r, testColor.g, testColor.b, swatchSolidLeftX, swatchSolidBottomY, swatchColor.r, swatchColor.g, swatchColor.b, x, y);
            }
            bool lowerRightResult = pixels.DoesPixelMatch(swatchSolidRightX, swatchSolidBottomY, swatchColor.IsMatch);
            if (!lowerRightResult)
            {
                pixels.ColorAt(swatchSolidRightX, swatchSolidBottomY, ref testColor);
                WriteLog(LogVerbosity.Excessive, "Lower-right mismatch for {8}, {9} - found {0},{1},{2} at {3}, {4} expected {5},{6},{7}", testColor.r, testColor.g, testColor.b, swatchSolidRightX, swatchSolidBottomY, swatchColor.r, swatchColor.g, swatchColor.b, x, y);
            }

            result &= upperRightResult;
            result &= lowerLeftResult;
            result &= lowerRightResult;
            if (!result)
            {
                // Box corners test failed
                WriteLog(LogVerbosity.High, "Failed to find left edge for potential swatch of color {2}, {3}, {4} at {0}, {1}", x, y, swatchColor.r, swatchColor.g, swatchColor.b);
                return false;
            }

            // scan down the right and left sides
            for (int yOff = 1; yOff < (swatchSolidHeight - 1); ++yOff)
            {
                result &= pixels.DoesPixelMatch(swatchSolidLeftX, swatchSolidTopY + yOff, swatchColor.IsMatch);
                if (!result)
                {
                    pixels.ColorAt(swatchSolidLeftX, swatchSolidTopY + yOff, ref testColor);
                    break;
                }
                result &= pixels.DoesPixelMatch(swatchSolidRightX, swatchSolidTopY + yOff, swatchColor.IsMatch);
                if (!result)
                {
                    pixels.ColorAt(swatchSolidRightX, swatchSolidTopY + yOff, ref testColor);
                }
            }
            if (!result)
            {
                WriteLog(LogVerbosity.Normal, "Failed to find left/right edges for potential swatch of color {2}, {3}, {4} at {0}, {1} [failed color = {5},{6},{7}]", x, y, swatchColor.r, swatchColor.g, swatchColor.b, testColor.r, testColor.g, testColor.b);
                return false;
            }
            for (int xOff = 1; xOff < (swatchSolidWidth - 1); ++xOff)
            {
                result &= pixels.DoesPixelMatch(swatchSolidLeftX + xOff, swatchSolidTopY, swatchColor.IsMatch);
                if (!result)
                {
                    pixels.ColorAt(swatchSolidLeftX + xOff, swatchSolidTopY, ref testColor);
                    break;
                }
                result &= pixels.DoesPixelMatch(swatchSolidLeftX + xOff, swatchSolidBottomY, swatchColor.IsMatch);
                if (!result)
                {
                    pixels.ColorAt(swatchSolidLeftX + xOff, swatchSolidBottomY, ref testColor);
                    break;
                }
            }
            if (!result)
            {
                WriteLog(LogVerbosity.Normal, "Failed to match upper/lower edges for potential swatch of color {2}, {3}, {4} at {0}, {1} [failed color = {5},{6},{7}]", x, y, swatchColor.r, swatchColor.g, swatchColor.b, testColor.r, testColor.g, testColor.b);
                return false;
            }


            // test the left edge for dark pixels -- the bottom-most pixel is bright now
            int i = 0;
            for (i = 1; result && i < swatchHeight - 1; ++i)
            {
                result &= pixels.DoesPixelMatch(x, y + i, PixelColor.IsDark);
            }
            if (!result)
            {
                // No dark border on the left side
                WriteLog(LogVerbosity.Normal, "Failed to find left border for potential swatch of color {2}, {3}, {4} at {0}, {1}", x, y, swatchColor.r, swatchColor.g, swatchColor.b);
                return false;
            }

            // test the dark top border and for papyrus above and below the swatch
            bool borderError = false;
            int papyErrorCount = 0;
            for (i = 0; result && (i < swatchWidth - 1); ++i)
            {
                bool isBorder = pixels.DoesPixelMatch(x + i, y, PixelColor.IsDark);
                result &= isBorder;
                if (!isBorder)
                {
                    WriteLog(LogVerbosity.Normal, "Probable swatch at {0},{1} failed upper border test at {2},{3}", x, y, x + i, y);
                    borderError = true;
                }

                // Checking along the top of the swatch for papyrus
                // The row just above is shaded, so check 2 above
                if (y > 1)
                {
                    bool isPapyrus = pixels.DoesPixelMatch(x + i, y - 2, PixelColor.IsPapyrus);
                    papyErrorCount += isPapyrus ? 0 : 1;

                }
                else
                {
                    ++papyErrorCount;
                }

                // Checking along the bottom of the swatch for papyrus
                if (y < pixels.height)
                {
                    bool isPapyrus = pixels.DoesPixelMatch(x + i, y + swatchHeight, PixelColor.IsPapyrus);
                    papyErrorCount += isPapyrus ? 0 : 1;
                }
                else
                {
                    ++papyErrorCount;
                }
            }

            result &= (papyErrorCount < (swatchWidth / 20)); // allow up to 5% error rate checking for papy texture, because this seems to be inconsistent
            if (!result && ((i > (swatchWidth*0.8)) || (papyErrorCount >= (swatchWidth/20))))
            {
                if (!borderError && (papyErrorCount < swatchWidth))
                {
                    WriteLog(LogVerbosity.Normal, "Found a potential swatch candidate of width {0} at {1},{2} that had {3} failures matching papyrus texture", i, x, y, papyErrorCount);
                }
            }

            return result;
        }

        unsafe private bool TestPosition(int x, int y, Pixels pixels, ref PaintColor reactedColor, ref int redPixelStart)
        {
            PixelColor pixelColor = new PixelColor();

            // Check 4 corners of solid area and left/right solid bar areas
            bool foundSwatch = IsPossibleSwatchUpperLeft(pixels, x, y); // ((pixel_r < 0x46) && (pixel_g < 0x46) && (pixel_b < 0x46));
            if (foundSwatch)
            {
                WriteLog(LogVerbosity.Normal, "Found probable swatch at {0},{1} - checking border slices", x, y);
                int borderXOffset = 0;
                for (borderXOffset = 2; foundSwatch && (borderXOffset < swatchTestWidth); ++borderXOffset)
                {
                    foundSwatch &= IsPossibleSwatchSlice(pixels, x + borderXOffset, y);
                    if (!foundSwatch)
                    {
                        WriteLog(LogVerbosity.Normal, "Failed slice test at {0},{1}", x + borderXOffset, y);
                        break;
                    }
                    foundSwatch &= IsPossibleSwatchSlice(pixels, x + swatchWidth - borderXOffset, y);
                    if (!foundSwatch)
                    {
                        WriteLog(LogVerbosity.Normal, "Failed slice test at {0},{1}", x + swatchWidth - borderXOffset, y);
                        break;
                    }
                }
            }

            if (foundSwatch)
            {
                // WE FOUND THE SWATCH!
                // Now we know where the color bars are.
                int redPixelCount = pixels.LengthOfColorAt(x, y + redBarSpacing, PixelColor.IsRed);
                reactedColor.Red = (byte)Math.Round((float)redPixelCount * 255f / (float)colorBarWidth);

                int greenPixelCount = pixels.LengthOfColorAt(x, y + greenBarSpacing, PixelColor.IsGreen);
                reactedColor.Green = (byte)Math.Round((float)greenPixelCount * 255f / (float)colorBarWidth);

                int bluePixelCount = pixels.LengthOfColorAt(x, y + blueBarSpacing, PixelColor.IsBlue);
                reactedColor.Blue = (byte)Math.Round((float)bluePixelCount * 255f / (float)colorBarWidth);
                WriteLog(LogVerbosity.Low, "Found the color swatch at {0}, {1}. Color={2} Red={3}px Green={4}px Blue={5}px", x, y, reactedColor, redPixelCount, greenPixelCount, bluePixelCount);
                return true;
            }
            return false;
        }

        unsafe public bool CaptureReaction(byte* pixBytes, int screenshotWidth, int screenshotHeight, int stride, int numChannels, int bitsPerSample)
        {
            PaintColor reactedColor = new PaintColor();
            int redPixelStart = 0;
            ScreenWidth = screenshotWidth;
            ScreenHeight = screenshotHeight;
            IsCaptured = false;
            _recordedColor.Clear();

            Pixels pixels = new Pixels
            {
                pixBytes = pixBytes,
                width = screenshotWidth,
                height = screenshotHeight,
                stride = stride,
                numChannels = numChannels,
                pixelMultiplier = pixelMultiplier
            };
            IsCaptured = false;
            if (!firstRun)
            {
                // If this is not the first run, let's check the last location, to see if the UI is still there.
                if (TestPosition(lastSwatchX, lastSwatchY, pixels, ref reactedColor, ref redPixelStart))
                {
                    IsCaptured = true;
                    RedBarX = (redPixelStart % stride) / numChannels;
                    RedBarY = redPixelStart / stride;
                    RecordedColor = reactedColor;
                    return true;
                }
                else
                {
                    firstRun = true;
                }
            }

            int verbosityIdx;
            AppSettings.Get("Log.Verbosity", out verbosityIdx, 1);
            logVerbosity = (LogVerbosity)verbosityIdx;

            int startX;
            int endX;
            int startY;
            int endY;
            AppSettings.Get("ScanArea.Min.X", out startX, 0);
            AppSettings.Get("ScanArea.Min.Y", out startY, 0);
            AppSettings.Get("ScanArea.Max.X", out endX, screenshotWidth);
            AppSettings.Get("ScanArea.Max.Y", out endY, screenshotHeight);
            startX = Math.Max(2, Math.Min(startX, screenshotWidth-2));
            startY = Math.Max(2, Math.Min(startY, screenshotHeight-2));
            endX = Math.Min(screenshotWidth-2, Math.Max(2, endX));
            endY = Math.Min(screenshotHeight-2, Math.Max(2, endY));

            int patchTestSize = ((swatchHeight - 5) / 2) - 1;
            PixelColor patchColor = new PixelColor();
            for (int roughX = startX; roughX < endX - colorBarWidth + patchTestSize; roughX += patchTestSize)
            {
                for (int roughY = startY; roughY < (endY - (blueBarSpacing + 10) + patchTestSize  /*53*/); roughY += patchTestSize)
                {
                    OnCaptureProgress?.Invoke(roughX, roughY);
                    if (!pixels.IsSolidPatchAt(roughX, roughY, patchTestSize, patchTestSize)) continue;
                    pixels.ColorAt(roughX, roughY, ref patchColor);
                    WriteLog(LogVerbosity.Excessive, "Found a solid patch of {2},{3},{4} at {0}, {1}", roughX, roughY, patchColor.r, patchColor.g, patchColor.b);
                    for (X = Math.Max(0, roughX - patchTestSize); X < roughX; ++X)
                    {
                        for (Y = Math.Max(0, roughY - patchTestSize); Y < roughY; ++Y)
                        {
                            WriteLog(LogVerbosity.Excessive, "Searching for potential swatch at {0},{1} after found square at {2},{3}", X, Y, roughX, roughY);
                            if (TestPosition(X, Y, pixels, ref reactedColor, ref redPixelStart))
                            {
                                RedBarX = (redPixelStart % stride) / numChannels;
                                RedBarY = redPixelStart / stride;
                                RecordedColor = reactedColor;
                                lastSwatchX = X;
                                lastSwatchY = Y;
                                firstRun = false;
                                IsCaptured = true;
                                return true;
                            }
                        }
                    }
                    WriteLog(LogVerbosity.Excessive, "False-positive patch of color {0},{1},{2} at {3},{4}", patchColor.r, patchColor.g, patchColor.b, roughX, roughY);
                }
            }
            return false;
        }

        public bool RecordReaction(PlayerProfile profile, PaintColor expectedColor, PaintColor reactedColor, Reagent[] reagents)
        {
            bool saved = false;
            int r, g, b;
            if (reagents[2] != null)
            {
                // A 3-reagent reaction.
                Reaction reaction1 = profile.FindReaction(reagents[0], reagents[1]);
                Reaction reaction2 = profile.FindReaction(reagents[0], reagents[2]);
                Reaction reaction3 = profile.FindReaction(reagents[1], reagents[2]);
                
                r = reactedColor.Red - expectedColor.Red;
                g = reactedColor.Green - expectedColor.Green;
                b = reactedColor.Blue - expectedColor.Blue;
                
                if (reaction2 == null)
                {
                    r = r - reaction1.Red - reaction3.Red;
                    g = g - reaction1.Green - reaction3.Green;
                    b = b - reaction1.Blue - reaction3.Blue;
                    Reaction reaction = new Reaction(r, g, b);
                    profile.SetReaction(reagents[0], reagents[2], reaction);
                    profile.Save();
                    saved = true;
                    OnReactionRecorded?.Invoke(reagents[0], reagents[2], reaction);
                }
                else if (reaction3 == null)
                {
                    r = r - reaction1.Red - reaction2.Red;
                    g = g - reaction1.Green - reaction2.Green;
                    b = b - reaction1.Blue - reaction2.Blue;
                    Reaction reaction = new Reaction(r, g, b);
                    profile.SetReaction(reagents[1], reagents[2], reaction);
                    profile.Save();
                    saved = true;
                    OnReactionRecorded?.Invoke(reagents[1], reagents[2], reaction);
                }
            }
            else if ((reagents[0] != null) && (reagents[1] != null))
            {
                // A 2-reagent reaction.
                r = reactedColor.Red - expectedColor.Red;
                g = reactedColor.Green - expectedColor.Green;
                b = reactedColor.Blue - expectedColor.Blue;
                Reaction reaction = new Reaction(r, g, b);
                profile.SetReaction(reagents[0], reagents[1], reaction);
                profile.Save();
                saved = true;
                OnReactionRecorded?.Invoke(reagents[0], reagents[1], reaction);
            }
            return saved;
        }
    }
}

