﻿/*
 * Copyright (c) 2010, Tess Snider

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
*/

using System;
using System.Collections.Generic;


namespace DesertPaintLab
{
    [System.ComponentModel.ToolboxItem(true)]
    public partial class SimulatorView : Gtk.Bin
    {
        Gtk.ListStore recipeData = new Gtk.ListStore(typeof(string), typeof(int));
        PaintRecipe paintRecipe;

        List<IngredientPair> missingWarned = new List<IngredientPair>();
        List<KeyValuePair<string, string>> newMissing = new List<KeyValuePair<string, string>>();

        private class IngredientPair
        {
            public string first;
            public string second;

            public IngredientPair(string first, string second)
            {
                this.first = first;
                this.second = second;
            }
        }

        public PlayerProfile Profile
        {
            set {
                paintRecipe.Clear();
                paintRecipe.Reactions = value.Reactions;
                UpdateRecipeColor();
            }
        }

        public SimulatorView(PlayerProfile profile) : base()
        {
            this.Build ();

            paintRecipe = new PaintRecipe();
            paintRecipe.Reactions = profile.Reactions;
            
            Gtk.TreeViewColumn reagentColumn = new Gtk.TreeViewColumn();
            Gtk.CellRendererText reagentColumnCell = new Gtk.CellRendererText();
            reagentColumn.PackStart(reagentColumnCell, true);       
            reagentColumn.Title = "Ingredient";
            
            reagentListView.AppendColumn(reagentColumn);
            reagentColumn.AddAttribute(reagentColumnCell, "text", 0);
            
            reagentListView.Model = ReagentManager.NameListModel;
            
            Gtk.TreeViewColumn additiveColumn = new Gtk.TreeViewColumn();
            Gtk.CellRendererText additiveColumnCell = new Gtk.CellRendererText();
            additiveColumn.PackStart(additiveColumnCell, true);     
            additiveColumn.Title = "Ingredient";
            
            recipeView.AppendColumn(additiveColumn);
            additiveColumn.AddAttribute(additiveColumnCell, "text", 0);
            
            Gtk.TreeViewColumn qtyColumn = new Gtk.TreeViewColumn();
            Gtk.CellRendererText qtyColumnCell = new Gtk.CellRendererText();
            qtyColumnCell.Editable = true;
            qtyColumnCell.Edited += OnQtyEdited;
            qtyColumn.PackStart(qtyColumnCell, true);       
            qtyColumn.Title = "Qty";
            
            recipeView.AppendColumn(qtyColumn);
            qtyColumn.AddAttribute(qtyColumnCell, "text", 1);
            
            recipeView.Model = recipeData;
            
            recipeData.RowChanged += OnRecipeChanged;
            recipeData.RowDeleted += OnRecipeChanged;
            recipeData.RowInserted += OnRecipeChanged;
            recipeData.RowsReordered += OnRecipeChanged;
            
        }
        
        protected virtual void OnAddReagent(object sender, System.EventArgs e)
        {
            Gtk.TreeModel model;
            Gtk.TreeIter iter;

            Gtk.TreeSelection selection = reagentListView.Selection;
            if ((selection != null) && selection.GetSelected(out model, out iter))
            {                   
                recipeData.AppendValues(model.GetValue(iter, 0).ToString(), 1); 
                
                recipeData.IterNthChild(out iter, recipeView.Children.Length - 1);
                
                selection = recipeView.Selection;
                selection.SelectIter(iter);
            } 
        }   
        
        protected void OnQtyEdited(object sender, Gtk.EditedArgs args)
        {
            Gtk.TreeIter iter;
            recipeData.GetIter(out iter, new Gtk.TreePath(args.Path));
            
            int oldValue = (int)recipeData.GetValue(iter, 1);
            
            try
            {
                recipeData.SetValue(iter, 1, int.Parse(args.NewText));
                UpdateRecipeColor();
            }
            catch (Exception)
            {
                recipeData.SetValue(iter, 1, oldValue);
            }
        }   
    
        protected void OnRecipeChanged(object sender, GLib.SignalArgs args)
        {
            UpdateRecipeColor();
        }   

        void UpdateRecipeColor()
        {
            if (recipeView.Children.Length == 0)
            {
                paintSwatch.Clear();
            }
            
            paintRecipe.Clear();

            Gtk.TreeIter iter;
            string reagentName;
            int qty;
            
            recipeData.GetIterFirst(out iter);
            
            do
            {
                reagentName = (string) recipeData.GetValue(iter, 0);
                
                if (reagentName == null)
                {
                    continue;   
                }
                
                qty = (int)recipeData.GetValue(iter, 1);
                for (int i = 0; i < qty; ++i)
                {
                    paintRecipe.AddReagent(reagentName);
                }
            }
            while (recipeData.IterNext(ref iter));
            
            PaintColor resultColor = new PaintColor(paintRecipe.ReactedColor);
            paintSwatch.Color = resultColor;
            if (paintRecipe.CheckMissingReactions(ref newMissing))
            {
                string warningMsg = "";

                foreach (KeyValuePair<string, string> newEntry in newMissing)
                {
                    IngredientPair match = missingWarned.Find(x => (x.first.Equals(newEntry.Key) && x.second.Equals(newEntry.Value)));
                    if (match == null)
                    {
                        match = new IngredientPair(newEntry.Key, newEntry.Value);
                        missingWarned.Add(match);
                        warningMsg += newEntry.Key + " + " + newEntry.Value + "\n";
                    }
                }
                if (warningMsg.Length > 0)
                {
                    Gtk.MessageDialog md = new Gtk.MessageDialog((Gtk.Window)Toplevel,
                        Gtk.DialogFlags.DestroyWithParent,
                        Gtk.MessageType.Warning,
                        Gtk.ButtonsType.Ok,
                        "These combinations have not yet had reactions recorded:\n\n" +
                        warningMsg);
                    md.Run();
                    md.Destroy();
                }
            }
            
        }
        
        protected virtual void OnIncrementReagent (object sender, System.EventArgs e)
        {
            Gtk.TreeModel model;
            Gtk.TreeIter iter;

            Gtk.TreeSelection selection = recipeView.Selection;
            if ((selection != null) && selection.GetSelected(out model, out iter))
            {
                int oldValue = (int)recipeData.GetValue(iter, 1);
                recipeData.SetValue(iter, 1, oldValue + 1);
            }
            
        }
        
        protected virtual void OnDecrementReagent (object sender, System.EventArgs e)
        {
            Gtk.TreeModel model;
            Gtk.TreeIter iter;

            Gtk.TreeSelection selection = recipeView.Selection;
            if ((selection != null) && selection.GetSelected(out model, out iter))
            {
                int oldValue = (int)recipeData.GetValue(iter, 1);
                if (oldValue == 1)
                {
                    recipeData.Remove(ref iter);
                }
                else
                {
                    recipeData.SetValue(iter, 1, oldValue - 1);
                }
            }           
            
        }

        protected virtual void OnFlushReagents (object sender, System.EventArgs e)
        {
            recipeData.Clear();
        }
    }
}